#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  3 13:26:27 2018

@author: Aishwarya Selvaraj


A .pyx file is compiled by cython to a .c file containing the code of a python
extension module.
This .c file is then compiled by c compiler to .so file which could be
imported directly into a python session.
The .pxd file is imported into .pyx file using the keyword cimport
"""
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
from atcore_h cimport *
from libc.stdlib cimport malloc, calloc, free

cimport cython
cimport numpy as np

import numpy as np
from atcore_defs import *
import matplotlib.pyplot as plt
from threading import Thread

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
cdef copy2WCStr(char *string, AT_WC *stringWC):
	cdef int i

	i = 0
	for c in string:
		stringWC[i] = <AT_WC> c
		i += 1
	stringWC[i] = '\0'

def convert2String(charlist):
	string = ""
	for c in charlist:
		try:
			if c != 0:
				string += chr(c)
		except:
			pass  # print c
	return string
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


cdef class andorCameraSDK3(object):
	cdef:
		AT_H cameraHandle
		bint set16bit
		double exposureTime
		AT_64 aoiHeight
		AT_64 aoiWidth
		AT_64 aoiStride
		AT_64 framecount
		double framerate
		double readoutime
		int hbin
		int vbin
		int AOITop
		int AOILeft



	#-------------------------------------------------------------------------------------------------------------------
	def __cinit__(self,master):
		self.cameraHandle = <AT_H>0
		self.set16bit = False
		self.exposureTime = 0
		self.aoiHeight = 0
		self.aoiWidth = 0
		self.aoiStride = 0
		self.framecount = 0
		self.framerate = 0
		self.readoutime = 0
		self.hbin =0
		self.vbin = 0
		self.AOITop =  0
		self.AOILeft = 0
		self.master = master



		cdef AT_H cameraHandle = -1

		# print "Opening Camera...",

		response_code = AT_Open(<int>0, &cameraHandle)
		if response_code != 0:
			pass
			#return response_code
			# print " FAILED: " + errorCodes[response_code]
		else:
			self.cameraHandle = cameraHandle
			# print " SUCCESS"

	#-------------------------------------------------------------------------------------------------------------------
	@classmethod
	def initializeCameraLibrary(self):
		response_code = AT_InitialiseLibrary()
		print "Opening AT Library..."
		return response_code
	#    if response_code != 0:
	#        # print " FAILED" + errorCodes[response_code]
	#    else:
	#        # print " SUCCESS"

	#-------------------------------------------------------------------------------------------------------------------
	@classmethod
	def closeCameraLibrary(classArg):
		response_code = AT_FinaliseLibrary()

		# print "Closing AT Library...",
		return response_code
	#    if response_code != 0:
	#        # print " FAILED" + errorCodes[response_code]
	#    else:
	#        # print " SUCCESS"

	#-------------------------------------------------------------------------------------------------------------------
	@classmethod
	def getCameraList(classArg):
		cdef long long deviceCount = 0
		cdef AT_WC string[15]
		cdef AT_WC *testString
		cdef AT_H HANDLE_SYSTEM = 1

		# print "Getting Number of Devices...",

		copy2WCStr("Device Count", string)

		response_code = AT_GetInt(HANDLE_SYSTEM, string, &deviceCount)

		return deviceCount


	#-------------------------------------------------------------------------------------------------------------------
#    def __dealloc__(self):
#        self.closeCamera()

	#-------------------------------------------------------------------------------------------------------------------
	def openCamera(self):
		cdef AT_H cameraHandle = -1

		# print "Opening Camera...",

		response_code = AT_Open(<int>0, &cameraHandle)
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			# print " SUCCESS"
			self.cameraHandle = cameraHandle
			return cameraHandle

	#-------------------------------------------------------------------------------------------------------------------
	def closeCamera(self):
		# print "Closing Camera...",

		response_code = AT_Close(self.cameraHandle)
		return response_code
#        if response_code != 0:
#            # print " FAILED" + errorCodes[response_code]
#        else:
#            # print " SUCCESS"

	#-------------------------------------------------------------------------------------------------------------------
	def getCameraString(self):
		cdef:
			AT_WC cameraModel[128]
			AT_WC string[15]
			int charstr_bytes
			char* charstr
			int copied_bytes

		print "Getting Camera String...",

		copy2WCStr("Camera Model",string)
		response_code = AT_GetString(<AT_H>self.cameraHandle, string, cameraModel, 128)

		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			print " SUCCESS"
			return cameraModel
	#-------------------------------------------------------------------------------------------------------------------
	def set_ExposureTime(self, exposureTime):
		cdef AT_WC string[15]

		print "Setting Exposure Time to...",
		copy2WCStr("Exposure Time",string)
		response_code = AT_SetFloat(<AT_H>self.cameraHandle, string, exposureTime);

		if response_code != 0:
			pass
			# print " FAILED: " + errorCodes[response_code]
		else:
			print exposureTime
			self.exposureTime = exposureTime
			#print self.exposureTime
		return response_code

	#-------------------------------------------------------------------------------------------------------------------
	def set_PixelReadoutRate(self,pixelreadout_type):
		cdef:
			AT_WC string[16]
			AT_WC string1[15]
			int available = 0

		print "Setting Pixel Readout Rate  ...",
		copy2WCStr("PixelReadoutRate",string)
		copy2WCStr(pixelreadout_type,string1)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string1)
			if response_code != 0:
				#print " FAILED "
				return response_code
			else:
				print "SUCCESS"
			#self.readoutrate = readoutrate


	#-------------------------------------------------------------------------------------------------------------------
	def get_ActualExposureTime(self):
		cdef:
			int available
			double exposureTime = 0
			AT_WC string[22]

		print "Getting Actual Exposure Time...",
		copy2WCStr("Actual Exposure Time",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available);
		if available:
			response_code = AT_GetFloat(<AT_H>self.cameraHandle, string, &exposureTime)
			if response_code != 0:
				# print " FAILED: " + errorCodes[response_code]
				return response_code
			else:
				print exposureTime
				#self.exposureTime = exposureTime
				return exposureTime
		else:
			pass
			# print "NOT AVAILABLE"
	#-----------------------------------------------------------------------------------------------------------------
	def get_PixelReadoutRate(self):
		cdef:
			AT_WC string[16]
			int i_index = 0
			AT_WC szValue[64]

		copy2WCStr("PixelReadoutRate",string)
		response_code = AT_GetEnumerated(<AT_H>self.cameraHandle, string, &i_index)
		if response_code != 0:
			#print " FAILED: "
			return response_code
		else:
			pass
			#print "SUCCESS:"
			#return response_code

		copy2WCStr("PixelReadoutRate",string)
		response_code = AT_GetEnumeratedString(<AT_H>self.cameraHandle, string, i_index, szValue, 64)
		if response_code != 0:
			print " FAILED: "
			#return response_code
		else:
			#print "SUCCESS"
			print "The Pixel Readout Rate is set to :",
			print convert2String(szValue)
		return szValue

	#-----------------------------------------------------------------------------------------------------------------
	def set_HeightWidth(self, height=0, width=0):
		cdef:
			int available = 0
			AT_64 aoiHeight
			AT_64 aoiWidth
			AT_WC string[15]

		aoiHeight = <AT_64> height
		aoiWidth = <AT_64> width

		########### CAMERA HEIGHT SETTING #################################

		copy2WCStr("AOI Height",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				if aoiHeight == 0:
					# print "Getting MAX Image Height...",
					response_code = AT_GetIntMax(<AT_H>self.cameraHandle, string, &aoiHeight);
					if response_code != 0:
						# print " FAILED: " + errorCodes[response_code]
						return response_code
					else:
						pass
						# print " SUCCESS"
				else:
					# print "Setting Camera Height to %d ..." %(aoiHeight)
					response_code = AT_SetInt(<AT_H>self.cameraHandle, string, aoiHeight);
					if response_code != 0:
						# print " FAILED: " + errorCodes[response_code]
						return response_code
					else:
						pass
						# print " SUCCESS"
			else:
				# print "Height setting NOT WRITABLE"
				# print "Getting MAX Image Height...",
				response_code = AT_GetIntMax(<AT_H>self.cameraHandle, string, &aoiHeight);
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					pass
					# print " SUCCESS: %d"%(aoiHeight)
		else:
			pass
			# print "Height setting NOT AVAILABLE"

		################# CAMERA WIDTH SETTING #############################

		copy2WCStr("AOI Width",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				if aoiWidth == 0:
					# print "Getting MAX Image Width...",
					response_code = AT_GetIntMax(<AT_H>self.cameraHandle, string, &aoiWidth);
					if response_code != 0:
						# print " FAILED: " + errorCodes[response_code]
						return response_code
					else:
						pass
						# print " SUCCESS"
				else:
					# print "Setting Camera Width to %d ..." %(aoiWidth),
					response_code = AT_SetInt(<AT_H>self.cameraHandle, string, aoiWidth);
					if response_code != 0:
						# print " FAILED: " + errorCodes[response_code]
						return response_code
					else:
						pass
						# print " SUCCESS"
			else:
				# print "Width setting NOT WRITABLE"
				# print "Getting MAX Image Width...",
				response_code = AT_GetIntMax(<AT_H>self.cameraHandle, string, &aoiWidth);
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					pass
					# print " SUCCESS: %d"%(aoiWidth)
		else:
			pass
			# print "Width setting NOT AVAILABLE"

		#print "The Image Size is :",
		#print (aoiHeight, aoiWidth)
		return (aoiHeight, aoiWidth)

	#-------------------------------------------------------------------------------------------------------------------
	def set_PreAmpGain(self):
		cdef:
			AT_WC string[25]
			AT_WC string2[40]
			int available

		copy2WCStr("SimplePreAmpGainControl",string)
		copy2WCStr("16-bit (low noise & high well capacity)",string2)
		"""
		Other options:
		11-bit (high well capacity)
		12-bit (high well capacity)
		11-bit (low noise)
		12-bit (low noise)
		16-bit (low noise & high well capacity)
		"""
		#print "Setting Preamp Gain Control... "
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available);


		set16bit = False
		self.set16bit = False

		if available:
			# print "AVAILABLE"
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			# print "Setting Preamp Gain Control...",
			if available:
				response_code=AT_SetEnumString(<AT_H>self.cameraHandle, string, string2)
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					set16bit = True
					self.set16bit = True
					print "PreAmp Gain is set to :",
					print convert2String(string2)
					return string2
		else:
			pass


	#-------------------------------------------------------------------------------------------------------------------
	def set_PixelEncoding(self):
		cdef:
			int available
			AT_WC string[15]
			AT_WC string2[15]

		#print "Setting Pixel Encoding..."
		copy2WCStr("PixelEncoding",string)

		if self.set16bit:
			copy2WCStr("Mono16",string2)
		else:
			copy2WCStr("Mono12Packed",string2)

		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			## print "AVAILABLE, ",
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				## print "WRITABLE, ",
				response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string2)
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					print "Pixel encoding is set to :",
					print convert2String(string2)
					return string2


	#-------------------------------------------------------------------------------------------------------------------
	def set_TriggerMode(self):
		cdef:
			int available
			AT_WC string[15]
			AT_WC string1[15]

		copy2WCStr("TriggerMode",string)
		copy2WCStr("Internal",string1)
		#print "In Trigger Mode"

		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			#print "AVAILABLE, "
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				#print "WRITABLE, ",
				response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string1)
				if response_code != 0:
					#print " FAILED: "
					return response_code
				else:
					#print "SUCCESS"
					print "The Trigger Mode is set to :",
					print convert2String(string1)
					return string1

	#-------------------------------------------------------------------------------------------------------------------
	def set_CycleMode(self):
		cdef:
			int available
			AT_WC string[15]
			AT_WC string1[5]

		copy2WCStr("CycleMode",string)
		copy2WCStr("Fixed",string1)
		print "In cycle mode"
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			#print "AVAILABLE, "
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				#print "WRITABLE, "
				response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string1)
				if response_code != 0:
					#print " FAILED: "
					return response_code
				else:
					#print " SUCCESS"
					return string1


	#-------------------------------------------------------------------------------------------------------------------
	def get_SensorDetails(self):
		cdef:
			int available
			AT_WC widthString[15]
			AT_WC heightString[15]
			AT_WC string[15]
			AT_64 aoiWidth
			AT_64 aoiHeight
			AT_64 aoiStride

		copy2WCStr("AOI Width",widthString)

		# print "Getting AOI Width...",
		AT_IsImplemented(<AT_H>self.cameraHandle, widthString, &available)
		if available == "AT_FALSE":
			# print "NOT AVAILABLE"
			# print "Getting Sensor Width...",
			copy2WCStr("Sensor Width",widthString)

		response_code = AT_GetInt(<AT_H>self.cameraHandle, widthString, &aoiWidth)
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			pass
			# print " SUCCESS"

		copy2WCStr("AOI Height",heightString)

		# print "Getting AOI Height...",
		AT_IsImplemented(<AT_H>self.cameraHandle, heightString, &available)
		if available == "AT_FALSE":
			# print "NOT AVAILABLE"
			# print "Getting Sensor Height...",
			copy2WCStr("Sensor Height",heightString)

		response_code = AT_GetInt(<AT_H>self.cameraHandle, heightString, &aoiHeight)
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			pass
			# print " SUCCESS"

		# print "Getting AOI Stride...",
		copy2WCStr("AOIStride",string)
		response_code = AT_GetInt(<AT_H>self.cameraHandle, string, &aoiStride)
		# print "AOI Stride = %d"%(aoiStride)
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			# print " SUCCESS"
			self.aoiWidth = aoiWidth
			self.aoiHeight = aoiHeight
			self.aoiStride = aoiStride
			return (widthString, heightString, aoiWidth, aoiHeight, aoiStride)

	#-------------------------------------------------------------------------------------------------------------------
	def set_AOI(self, AOI_Settings):
		cdef:
			int AOIHBin = AOI_Settings[0]
			int AOIVBin = AOI_Settings[1]
			AT_64 AOILeft = AOI_Settings[2]
			AT_64 AOIWidth = AOI_Settings[3]
			AT_64 AOITop = AOI_Settings[4]
			AT_64 AOIHeight = AOI_Settings[5]
			AT_WC string[15]
			AT_WC string1[20]
			AT_WC binningString[4]
			int available
			int temp = 0

		print "Setting AOI Binning...",

		copy2WCStr("AOIBinning",string)
		binningString[1] = 'x'
		if AOIHBin == 1:
			binningString[0] = '1'
			binningString[2] = '1'
		if AOIHBin == 2:
			binningString[0] = '2'
			binningString[2] = '2'
		elif AOIHBin == 3:
			binningString[0] = '3'
			binningString[2] = '3'
		elif AOIHBin == 4:
			binningString[0] = '4'
			binningString[2] = '4'
		elif AOIHBin == 8:
			binningString[0] = '8'
			binningString[2] = '8'
		binningString[3] = '\0'

		result = str(chr(binningString[0]))+str(chr(binningString[1]))+str(chr(binningString[2]))
		self.hbin = AOIHBin
		self.vbin = AOIVBin


		copy2WCStr(result,string1)


		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available);
		if available:
			response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string1)
			if response_code != 0:
				return response_code
			else:
				print " SUCCESS"
		else:
			pass


		#AT_GetEnumIndex(<AT_H>self.cameraHandle, string, &temp) # TO see which binning option was selected
		#print temp





		# Follow the same order: Height, Width, Left, Top. Otherwise, the SDK throws AT_OUT_OF_RANGE error.
		print "Setting AOI Height...",
		copy2WCStr("AOIHeight",string)
		response_code = AT_SetInt(<AT_H>self.cameraHandle, string, AOIHeight)
		if response_code != 0:
			print " FAILED: "
		else:
			print " SUCCESS",
			print AOIHeight

		print "Setting AOI Width...",
		copy2WCStr("AOIWidth",string)
		response_code = AT_SetInt(<AT_H>self.cameraHandle, string, AOIWidth)
		if response_code != 0:
			print " FAILED: "
		else:
			print " SUCCESS",
			print AOIWidth

		print "Setting AOI Left pixel...",
		copy2WCStr("AOILeft",string)
		response_code = AT_SetInt(<AT_H>self.cameraHandle, string, AOILeft)
		if response_code != 0:
			print " FAILED: "
		else:
			self.AOILeft = AOILeft
			print " SUCCESS",
			print AOILeft


		print "Setting AOI Top pixel...",
		copy2WCStr("AOITop",string)
		response_code = AT_SetInt(<AT_H>self.cameraHandle, string, AOITop)
		if response_code != 0:
			print " FAILED: "
		else:
			self.AOITop =  AOITop
			print " SUCCESS",
			print AOITop
	#-------------------------------------------------------------------------------------------------------------------
	#def createBuffer(self,acqBuffer):
		#cdef:
			#AT_64 sizeInBytes
			#AT_WC string[20]
			#unsigned char * acqBuffer = NULL

		#copy2WCStr("ImageSizeBytes",string)
		#response_code = AT_GetInt(<AT_H> self.cameraHandle, string, &sizeInBytes)

		#if (response_code == 0):
			#acqBuffer =  <unsigned char *>calloc(sizeInBytes+7, sizeof(unsigned char))
			#response_code = AT_QueueBuffer(<AT_H>self.cameraHandle, acqBuffer, sizeInBytes)
			#if (response_code != 0):
				#print " FAILED: "
				#return response_code
			#else:
				#print " SUCCESS :"
				#return acqBuffer
		#else:
			#pass

	#-------------------------------------------------------------------------------------------------------------------
	def set_FrameCount(self,accumulatecount,framecount):
		cdef:
			int available
			AT_WC string[15]
			AT_WC string1[15]


		copy2WCStr("AccumulateCount",string1)
		#print "In AccumulateCount Setting "
		AT_IsImplemented(<AT_H>self.cameraHandle, string1, &available)
		if available:
			## print "AVAILABLE, ",
			AT_IsWritable(<AT_H>self.cameraHandle, string1, &available)

			if available:
				## print "WRITABLE, ",
				response_code = AT_SetInt(<AT_H>self.cameraHandle, string1, accumulatecount)
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					pass
					#print " SUCCESS"


		copy2WCStr("FrameCount",string)
		print "Setting Frame Count... ",
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			## print "AVAILABLE, ",
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				#print "WRITABLE, ",
				response_code = AT_SetInt(<AT_H>self.cameraHandle, string,framecount)
				if response_code != 0:
					# print " FAILED: " + errorCodes[response_code]
					return response_code
				else:
					print "SUCCESS"
					self.framecount = framecount
					print "Accumulate count  = ",
					print accumulatecount,
					print "Frame Count = ",
					print framecount

	#-------------------------------------------------------------------------------------------------------------------
	def set_FrameRate(self,framerate):
		cdef:
			int available
			AT_WC string[15]


		copy2WCStr("FrameRate",string)
		print "In FrameRate Setting "
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		response_code = AT_SetFloat(<AT_H>self.cameraHandle, string, framerate)
		if response_code != 0:
			print " FAILED: "
			pass
		else:
			print " SUCCESS"
			self.framerate = framerate
		return response_code

	#-------------------------------------------------------------------------------------------------------------------
	def get_FrameRate(self):
		cdef:
			int available
			double framerate = 0
			AT_WC string[22]

		#print "Getting Actual Frame Rate...",
		copy2WCStr("FrameRate",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available);
		if available:
			response_code = AT_GetFloat(<AT_H>self.cameraHandle, string, &framerate)
			if response_code != 0:
				print "FAILED: "
			else:
				#print "SUCCESS"
				return framerate
		else:
			pass
			#print "NOT AVAILABLE"


	#--------------------------------------------------------------------------------------------------
	@cython.boundscheck(False)
	@cython.wraparound(False)
	def acquireImage(self, np.ndarray[np.uint16_t,ndim=2,mode='c']data):
		cdef:
			AT_64 sizeInBytes
			AT_64 aoiStride
			AT_WC string[20]
			unsigned char * puc_image
			unsigned char *_puc_image
			unsigned short *pus_image
			unsigned int timeOut
			unsigned char * returnBuffer = NULL
			int bufferSize = 0


		# Image Size in Bytes = Height x Stride
		#print "Getting Image Size Bytes...",

		copy2WCStr("ImageSizeBytes",string)
		AT_GetInt(<AT_H> self.cameraHandle, string, &sizeInBytes);
		#print "Image Size Bytes = %d"%(sizeInBytes)

		#print "Getting AOI Stride...",
		copy2WCStr("AOIStride",string)
		AT_GetInt(<AT_H> self.cameraHandle, string, &aoiStride)
		#print "AOI Stride = %d"%(aoiStride)

		puc_image  = <unsigned char *>calloc(sizeInBytes, sizeof(unsigned char))

		#print "Queueing buffer...",
		response_code = AT_QueueBuffer(<AT_H>self.cameraHandle, puc_image, sizeInBytes)
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			pass
			#print " SUCCESS"

		#print "Starting Acquisition...",

		copy2WCStr("Acquisition Start",string)
		response_code = AT_Command(<AT_H>self.cameraHandle, string);
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			pass
			#print " SUCCESS"



		#self.exposureTime = self.get_ActualExposureTime()
		#print "Exposure Time = %f" %self.exposureTime

		#print "Waiting for Buffer...",
		#timeOut = <unsigned int>(3 * self.get_ActualExposureTime() * 1000);

		#if timeOut < 500:
			#timeOut = 500
		#print timeOut

		response_code = AT_WaitBuffer(<AT_H>self.cameraHandle, &returnBuffer, &bufferSize, 1000);
		if response_code != 0:
			# print " FAILED: " + errorCodes[response_code]
			return response_code
		else:
			pass
			#print " SUCCESS"

		if (returnBuffer != puc_image):
			print "FAILED...Returned buffer not equal to queued buffer. Image Acquisition aborted"
			return False
		elif (bufferSize != sizeInBytes):
			print "Returned buffer size not correct :  Expected: %ld, Actual: %ld. Image Acquisition aborted."%(sizeInBytes, bufferSize)
			return False

		cdef unsigned int i,j


		#print "Verification ..."
		#print "Height :",
		#print (self.aoiHeight)
		#print "Width  :",
		#print (self.aoiWidth)

		try:
			_puc_image = puc_image
			pus_image = <unsigned short*>puc_image
			for i in range(self.aoiWidth/self.vbin):
				pus_image = <unsigned short*>(_puc_image);
				for j in range(self.aoiHeight/self.hbin):
					#print "[%d %d] = %d"%(i,j, pus_image[0])
					data[i][j] = pus_image[0] #& 0x07FF
					pus_image += 1
				#print "[%d %d] = %d"%(i,j, pus_image[0])
				_puc_image += aoiStride;
		except Exception, e:
			print "Exception in function: ",
			print e

		#print "Stopping Acquisition..."

		copy2WCStr("AcquisitionStop",string)
		AT_Command(<AT_H>self.cameraHandle, string)
		AT_Flush(<AT_H>self.cameraHandle);
		free(puc_image)
		#print self.master
		self.master.Plot()


		return data


	#-------------------------------------------------------------------------------------------------------------------
	@cython.boundscheck(False)
	@cython.wraparound(False)
	def LiveAquisition(self,nframes,np.ndarray[np.uint16_t,ndim = 3,mode = 'c']data):
		cdef:
			int available
			AT_64 sizeInBytes
			AT_64 aoiStride
			AT_WC string[20]
			AT_WC string1[20]
			AT_WC string2[20]
			AT_WC string3[20]
			unsigned char * pBuf
			unsigned char * _puc_image
			int BufSize
			unsigned int i, j, k, l = 0


		threads =[]
		image = np.zeros(((self.aoiWidth-self.AOILeft)/self.hbin, (self.aoiHeight-self.AOITop)/self.vbin), dtype=np.uint16)
		image = np.zeros(((self.aoiWidth)/self.hbin, (self.aoiHeight)/self.vbin), dtype=np.uint16)

		copy2WCStr("Acquisition Start",string)
		response_code = AT_Command(<AT_H>self.cameraHandle, string)
		if response_code!= 0:
			print " FAILED: "
			return response_code
		else:
			pass
			print "SUCCESS"

		copy2WCStr("ImageSizeBytes",string1)
		AT_GetInt(<AT_H> self.cameraHandle, string1, &sizeInBytes)

		copy2WCStr("AOIStride",string1)
		AT_GetInt(<AT_H> self.cameraHandle, string1, &aoiStride)

		for i in range(nframes):
			pBuf = <unsigned char *>calloc(sizeInBytes, sizeof(unsigned char))
			AT_QueueBuffer(<AT_H>self.cameraHandle, pBuf, sizeInBytes)

			#print "Frame number is :",
			#print i


			response_code = AT_WaitBuffer(<AT_H>self.cameraHandle, &pBuf, &BufSize, 500)
			if response_code!=0:
				print "ERROR :Acquisition timeout :"
				return response_code
			else:
				pass
				#print "SUCCESS"


			#if (i==0) or (i == 1) or (i ==2) or (i==3) or (i==4):
			process = Thread(target = self.StoreArray,args =[pBuf,image,i])
			process.start()
			threads.append(process)

			#if (i==0) or (i == 1) or (i ==2) or (i==3) or (i==4):
				#data[i][:][:] = self.StoreArray(pBuf,image)

			#if (i==0) or (i == 1) or (i ==2) or (i==3) or (i==4):
				#_puc_image = pBuf
				#pus_image = <unsigned short*>pBuf
				#for j in range(self.aoiWidth/self.hbin):
					#pus_image = <unsigned short*>(_puc_image)
					#for k in range(self.aoiHeight/self.vbin):
						#data[l][j][k]  = pus_image[0]
						#pus_image += 1
					#_puc_image += aoiStride
				#l = l+1


		for process in threads:
			process.join()

		copy2WCStr("AcquisitionStop",string3)
		AT_Command(<AT_H>self.cameraHandle, string3)

		AT_Flush(<AT_H>self.cameraHandle);
		print "Deleting the Buffer :"
		free(pBuf)
		return data

	#-------------------------------------------------------------------------------------------------------------------
	@cython.boundscheck(False)
	@cython.wraparound(False)
	def StoreArray(self,unsigned char * pBuf,image,i):
		cdef:
			unsigned char *_puc_image
			unsigned short *pus_image


		_puc_image = pBuf
		pus_image = <unsigned short*>pBuf
		for j in range(self.aoiWidth/self.hbin):
			pus_image = <unsigned short*>(_puc_image)
			for k in range(self.aoiHeight/self.vbin):
				image[j][k]  = pus_image[0]
				pus_image += 1
			_puc_image += self.aoiStride

		#import matplotlib.pyplot as plt
		#plt.figure(1)
		#plt.imshow(image)
		#plt.show()
		#return image

	#-------------------------------------------------------------------------------------------------------------------
	def get_ReadoutTime(self):
		cdef:
			int available
			double readoutime = 0
			AT_WC string[22]


		copy2WCStr("ReadoutTime",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available);
		if available:
			AT_IsReadable(<AT_H>self.cameraHandle, string, &available)
			if available:
				response_code = AT_GetFloat(<AT_H>self.cameraHandle, string, &readoutime)
				if response_code != 0:
					print " FAILED: "
					return response_code
				else:
					print " SUCCESS"
					return readoutime
		else:
			pass
			# print "NOT AVAILABLE"

	#-------------------------------------------------------------------------------------------------------------------
	def set_ShutterMode(self,shutter_type):
		cdef:
			int available
			AT_WC string[25]
			AT_WC string1[7]

		copy2WCStr("ElectronicShutteringMode",string)
		copy2WCStr(shutter_type,string1)
		#print "In Shutter Mode"

		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			#print "AVAILABLE, ",
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				#print "WRITABLE, ",
				response_code = AT_SetEnumString(<AT_H>self.cameraHandle, string, string1)
				if response_code != 0:
					#print " FAILED: "
					return response_code
				else:
					print "The Shutter Mode is set to:",
					print convert2String(string1)
					return string1

	#-------------------------------------------------------------------------------------------------------------------
	def set_FastAOIFrameRateEnable(self):
		cdef:
			int available
			AT_WC string[25]


		copy2WCStr("FastAOIFrameRateEnable",string)
		print "Setting FastAOI FrameRate :",

		response_code = AT_SetBool(<AT_H>self.cameraHandle, string, AT_TRUE)
		if response_code != 0:
			print " FAILED: "
		else:
			print "SUCCESS"

	#-------------------------------------------------------------------------------------------------------------------
	def set_VerticallyCentreAOI(self):
		cdef:
			int available
			AT_WC string[20]

		#print "In Setting Vertically Centred AOI"
		copy2WCStr("VerticallyCentreAOI",string)
		response_code = AT_SetBool(<AT_H>self.cameraHandle, string, AT_TRUE)
		if response_code != 0:
			print "FAILED: "
		else:
			print "SUCCESS"




	#-------------------------------------------------------------------------------------------------------------------
	def set_CameraTemperature(self, temperatureIndex):
		cdef:
			AT_WC string[20]
			int available

		print "Enabling Sensor cooling..."
		copy2WCStr("SensorCooling",string)
		response_code = AT_SetBool(<AT_H>self.cameraHandle, string, AT_TRUE)
		if response_code != 0:
			print " FAILED: "
		else:
			print " SUCCESS"

		print "Setting Temperature...",

		#copy2WCStr("SensorTemperature",string)
		copy2WCStr("TemperatureControl",string)
		AT_IsImplemented(<AT_H>self.cameraHandle, string, &available)
		if available:
			print "AVAILABLE, ",
			AT_IsWritable(<AT_H>self.cameraHandle, string, &available)

			if available:
				response_code = AT_SetEnumIndex(<AT_H>self.cameraHandle, string, temperatureIndex)
				# response_code = AT_SetFloat(<AT_H>self.cameraHandle, string, <double>temperature)
				if response_code != 0:
					print " FAILED: "
				else:
					print " SUCCESS"
			else:
				print "NOT WRITABLE"


	#-------------------------------------------------------------------------------------------------------------------
	def get_Temperature(self):
		cdef:
			double temperature = 9999.9999
			AT_WC string[20]

		copy2WCStr("SensorTemperature",string)
		response_code = AT_GetFloat(<AT_H>self.cameraHandle, string, &temperature)

		return temperature # Returns 9999.9999 for validation in Python in case the temperature could not be read.


	#-----------------------------------------------------------------------------------------------
	def get_TemperatureStatus(self):
		cdef:
			int temperatureStatusIndex = 0
			int temperatureCount = 0
			AT_WC temperatureStatus[256]
			AT_WC string[20]

		copy2WCStr("TemperatureStatus",string)
		AT_GetEnumIndex(<AT_H>self.cameraHandle, string, &temperatureStatusIndex)

		AT_GetEnumStringByIndex(<AT_H>self.cameraHandle, string, temperatureStatusIndex, temperatureStatus, 256)

		#print "Stabilizing Temperature..."
		#while (convert2String(temperatureStatus)!="Stabilised"):
			#AT_GetEnumCount(<AT_H>self.cameraHandle, string, &temperatureCount)


			#print "Temperature is = ",
			#print self.get_Temperature()

			#AT_GetEnumIndex(<AT_H>self.cameraHandle, string, &temperatureStatusIndex)
			#AT_GetEnumStringByIndex(<AT_H>self.cameraHandle, string, temperatureStatusIndex, temperatureStatus, 256);

		#print "Temperature Stabilized..."
		return temperatureStatus

	#-------------------------------------------------------------------------------------------------------------------
	def CircularBuffer(self):
		cdef:
			AT_64 sizeInBytes
			int BufSize
			int BufferSize
			int NumberOfBuffers = 10
			int NumberOfFrames = 100
			AT_WC string[20]
			AT_WC string1[20]
			AT_WC string2[20]
			AT_WC string3[20]
			AT_WC string4[20]
			unsigned char * pBuf
			unsigned char * AcqBuffers


		copy2WCStr("ImageSizeBytes",string)
		AT_GetInt(<AT_H> self.cameraHandle, string, &sizeInBytes)

		#for i in range(NumberOfBuffers):
		AcqBuffers = <unsigned char *>calloc(sizeInBytes, sizeof(unsigned char))

		#for i in range(NumberOfBuffers):
		AT_QueueBuffer(<AT_H>self.cameraHandle, AcqBuffers, sizeInBytes)

		copy2WCStr("CycleMode",string1)
		copy2WCStr("Continous",string2)
		AT_SetEnumString(<AT_H>self.cameraHandle,string1,string2)

		copy2WCStr("AcquisitionStart",string3)
		AT_Command(<AT_H>self.cameraHandle, string3)

		for i in range(NumberOfFrames):
			print i
			AT_WaitBuffer(<AT_H>self.cameraHandle, &pBuf, &BufSize, 500)
			AT_QueueBuffer(<AT_H>self.cameraHandle, AcqBuffers, sizeInBytes)

		copy2WCStr("AcquisitionStop",string4)
		AT_Command(<AT_H>self.cameraHandle, string4)
		AT_Flush(<AT_H>self.cameraHandle)
		print "Done"

		free(AcqBuffers)
